#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present <https://github.com/TelegramPlayGround>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.
from Badmunda import raw

from .upgraded_gift_attribute_id import UpgradedGiftAttributeId


class UpgradedGiftAttributeIdBackdrop(UpgradedGiftAttributeId):
    """Identifier of a gift backdrop.

    Parameters:
        backdrop_id (``int``):
            Identifier of the sticker representing the backdrop.
    """
    def __init__(
        self,
        backdrop_id: int,
    ):
        super().__init__()

        self.backdrop_id = backdrop_id

    def write(self) -> "raw.types.StarGiftAttributeIdBackdrop":
        return raw.types.StarGiftAttributeIdBackdrop(
            backdrop_id=self.backdrop_id
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from uuid import uuid4

import Badmunda
from Badmunda import types
from ..object import Object


class InlineQueryResult(Object):
    """One result of an inline query.

    - :obj:`~Badmunda.types.InlineQueryResultCachedAudio`
    - :obj:`~Badmunda.types.InlineQueryResultCachedDocument`
    - :obj:`~Badmunda.types.InlineQueryResultCachedAnimation`
    - :obj:`~Badmunda.types.InlineQueryResultCachedPhoto`
    - :obj:`~Badmunda.types.InlineQueryResultCachedSticker`
    - :obj:`~Badmunda.types.InlineQueryResultCachedVideo`
    - :obj:`~Badmunda.types.InlineQueryResultCachedVoice`
    - :obj:`~Badmunda.types.InlineQueryResultArticle`
    - :obj:`~Badmunda.types.InlineQueryResultAudio`
    - :obj:`~Badmunda.types.InlineQueryResultContact`
    - :obj:`~Badmunda.types.InlineQueryResultDocument`
    - :obj:`~Badmunda.types.InlineQueryResultAnimation`
    - :obj:`~Badmunda.types.InlineQueryResultLocation`
    - :obj:`~Badmunda.types.InlineQueryResultPhoto`
    - :obj:`~Badmunda.types.InlineQueryResultVenue`
    - :obj:`~Badmunda.types.InlineQueryResultVideo`
    - :obj:`~Badmunda.types.InlineQueryResultVoice`
    """

    def __init__(
        self,
        type: str,
        id: str,
        input_message_content: "types.InputMessageContent",
        reply_markup: "types.InlineKeyboardMarkup"
    ):
        super().__init__()

        self.type = type
        self.id = str(uuid4()) if id is None else str(id)
        self.input_message_content = input_message_content
        self.reply_markup = reply_markup

    async def write(self, client: "Badmunda.Client"):
        pass



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import Badmunda
from Badmunda import raw, types


class GetBusinessConnection:
    async def get_business_connection(
        self: "Badmunda.Client",
        business_connection_id: str
    ):
        """Use this method to get information about the connection of the bot with a business account.

        .. include:: /_includes/usable-by/bots.rst

        Parameters:
            connection_id (``str``):
                Unique identifier of the business connection.

        Returns:
            :obj:`~Badmunda.types.BusinessConnection`: On success the business connection is returned.

        Example:
            .. code-block:: python

                # Get a business connection information
                await app.get_business_connection(connection_id)
        """
        r = await self.invoke(
            raw.functions.account.GetBotBusinessConnection(
                connection_id=business_connection_id
            )
        )

        users = {i.id: i for i in r.users}

        return types.BusinessConnection._parse(self, r.updates[0].connection, users)



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Union

import Badmunda
from Badmunda import raw, types, enums


class SetPrivacy:
    async def set_privacy(
        self: "Badmunda.Client",
        key: "enums.PrivacyKey",
        rules: List[Union[
            "types.InputPrivacyRuleAllowAll",
            "types.InputPrivacyRuleAllowBots",
            "types.InputPrivacyRuleAllowChats",
            "types.InputPrivacyRuleAllowCloseFriends",
            "types.InputPrivacyRuleAllowContacts",
            "types.InputPrivacyRuleAllowPremium",
            "types.InputPrivacyRuleAllowUsers",
            "types.InputPrivacyRuleDisallowAll",
            "types.InputPrivacyRuleDisallowBots",
            "types.InputPrivacyRuleDisallowChats",
            "types.InputPrivacyRuleDisallowContacts",
            "types.InputPrivacyRuleDisallowUsers"
        ]],
    ):
        """Set account privacy rules.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            key (:obj:`~Badmunda.enums.PrivacyKey`):
                Privacy key.

            rules (Iterable of :obj:`~Badmunda.types.InputPrivacyRule`):
                List of privacy rules.

        Returns:
            List of :obj:`~Badmunda.types.PrivacyRule`: On success, the list of privacy rules is returned.

        Example:
            .. code-block:: python

                from Badmunda import enums, types

                # Prevent everyone from seeing your phone number
                await app.set_privacy(enums.PrivacyKey.PHONE_NUMBER, [types.InputPrivacyRuleDisallowAll()])
        """
        r = await self.invoke(
            raw.functions.account.SetPrivacy(
                key=key.value(),
                rules=[await rule.write(self) for rule in rules]
            )
        )

        users = {i.id: i for i in r.users}
        chats = {i.id: i for i in r.chats}

        return types.List(types.PrivacyRule._parse(self, rule, users, chats) for rule in r.rules)



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from typing import Optional, Union

from Badmunda import raw, types, utils

from ..object import Object


class SuccessfulPayment(Object):
    """A service message about a new successful payment.

    Parameters:
        currency (``str``):
            Three-letter ISO 4217 `currency <https://core.telegram.org/bots/payments#supported-currencies>`_ code, or ``XTR`` for payments in `Telegram Stars <https://t.me/BotNews/90>`_.

        total_amount (``int``):
            Total price in the smallest units of the currency (integer, **not** float/double). For example, for a price of ``US$ 1.45`` pass ``amount = 145``. See the __exp__ parameter in `currencies.json <https://core.telegram.org/bots/payments/currencies.json>`_, it shows the number of digits past the decimal point for each currency (2 for the majority of currencies).

        invoice_payload  (``str``, *optional*):
            Bot specified invoice payload. Only available to the bot that received the payment.

        telegram_payment_charge_id (``str``, *optional*):
            Telegram payment identifier. Only available to the bot that received the payment.

        provider_payment_charge_id (``str``, *optional*):
            Provider payment identifier. Only available to the bot that received the payment.

        shipping_option_id (``str``, *optional*):
            Identifier of the shipping option chosen by the user. Only available to the bot that received the payment.

        payment_info (:obj:`~Badmunda.types.PaymentInfo`, *optional*):
            Payment information provided by the user. Only available to the bot that received the payment.

        is_recurring (``bool``, *optional*):
            True, if the payment is a recurring payment for a subscription.

        is_first_recurring (``bool``, *optional*):
            True, if the payment is the first payment for a subscription.

        invoice_slug (``str``, *optional*):
            Name of the invoice.

        subscription_expiration_date (:py:obj:`~datetime.datetime`, *optional*):
            Expiration date of the subscription, in Unix time; for recurring payments only.
    """

    def __init__(
        self, *,
        currency: str,
        total_amount: str,
        invoice_payload: str,
        telegram_payment_charge_id: str,
        provider_payment_charge_id: str,
        shipping_option_id: Optional[str] = None,
        order_info: Optional["types.OrderInfo"] = None,
        is_recurring: Optional[bool] = None,
        is_first_recurring: Optional[bool] = None,
        invoice_slug: Optional[str] = None,
        subscription_expiration_date: datetime = None,
    ):
        super().__init__()

        self.currency = currency
        self.total_amount = total_amount
        self.invoice_payload = invoice_payload
        self.telegram_payment_charge_id = telegram_payment_charge_id
        self.provider_payment_charge_id = provider_payment_charge_id
        self.shipping_option_id = shipping_option_id
        self.order_info = order_info
        self.is_recurring = is_recurring
        self.is_first_recurring = is_first_recurring
        self.invoice_slug = invoice_slug
        self.subscription_expiration_date = subscription_expiration_date

    @staticmethod
    def _parse(
        payment: Union[
            "raw.types.MessageActionPaymentSent",
            "raw.types.MessageActionPaymentSentMe"
        ]) -> "SuccessfulPayment":
        invoice_payload = None
        telegram_payment_charge_id = None
        provider_payment_charge_id = None
        shipping_option_id = None
        order_info = None

        if isinstance(payment, raw.types.MessageActionPaymentSentMe):
            # Try to decode invoice payload into string. If that fails, fallback to bytes instead of decoding by
            # ignoring/replacing errors, this way, button clicks will still work.
            try:
                invoice_payload = payment.payload.decode()
            except (UnicodeDecodeError, AttributeError):
                invoice_payload = payment.payload

            telegram_payment_charge_id = payment.charge.id
            provider_payment_charge_id = payment.charge.provider_charge_id
            shipping_option_id = getattr(payment, "shipping_option_id")

            if payment.info:
                payment_info = payment.info

                order_info = types.OrderInfo(
                    name=getattr(payment_info, "name", None),
                    phone_number=getattr(payment_info, "phone", None),
                    email=getattr(payment_info, "email", None),
                    shipping_address=types.ShippingAddress._parse(
                        payment_info.shipping_address
                    )
                )

        return SuccessfulPayment(
            currency=payment.currency,
            total_amount=payment.total_amount,
            invoice_payload=invoice_payload,
            telegram_payment_charge_id=telegram_payment_charge_id,
            provider_payment_charge_id=provider_payment_charge_id,
            shipping_option_id=shipping_option_id,
            order_info=order_info,
            is_recurring=getattr(payment, "recurring_used", None),
            is_first_recurring=getattr(payment, "recurring_init", None),
            invoice_slug=getattr(payment, "invoice_slug", None),
            subscription_expiration_date=utils.timestamp_to_datetime(payment.subscription_until_date),
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import logging
from datetime import datetime
from typing import List, Optional, Union

import Badmunda
from Badmunda import enums, raw, types, utils

log = logging.getLogger(__name__)

class SendPoll:
    async def send_poll(
        self: "Badmunda.Client",
        chat_id: Union[int, str],
        question: str,
        options: List[str],
        is_anonymous: bool = True,
        type: "enums.PollType" = enums.PollType.REGULAR,
        allows_multiple_answers: Optional[bool] = None,
        correct_option_id: Optional[int] = None,
        question_parse_mode: Optional["enums.ParseMode"] = None,
        question_entities: Optional[List["types.MessageEntity"]] = None,
        explanation: Optional[str] = None,
        explanation_parse_mode: Optional["enums.ParseMode"] = None,
        explanation_entities: Optional[List["types.MessageEntity"]] = None,
        open_period: Optional[int] = None,
        close_date: Optional[datetime] = None,
        is_closed: Optional[bool] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        message_thread_id: Optional[int] = None,
        effect_id: Optional[int] = None,
        reply_parameters: Optional["types.ReplyParameters"] = None,
        schedule_date: Optional[datetime] = None,
        business_connection_id: Optional[str] = None,
        options_parse_mode: Optional["enums.ParseMode"] = None,
        allow_paid_broadcast: bool = None,
        paid_message_star_count: int = None,
        reply_markup: Optional[
            Union[
                "types.InlineKeyboardMarkup",
                "types.ReplyKeyboardMarkup",
                "types.ReplyKeyboardRemove",
                "types.ForceReply"
            ]
        ] = None,

        reply_to_message_id: Optional[int] = None,
        reply_to_chat_id: Optional[Union[int, str]] = None,
        quote_text: Optional[str] = None,
        quote_parse_mode: Optional["enums.ParseMode"] = None,
        quote_entities: Optional[List["types.MessageEntity"]] = None,
        quote_offset: Optional[int] = None,
    ) -> "types.Message":
        """Send a new poll.

        .. include:: /_includes/usable-by/users-bots.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".
                For a contact that exists in your Telegram address book you can use his phone number (str).

            question (``str``):
                Poll question, 1-255 characters.

            options (List of ``str``):
                List of answer options, 2-12 strings 1-100 characters each.

            is_anonymous (``bool``, *optional*):
                True, if the poll needs to be anonymous.
                Defaults to True.

            type (:obj`~Badmunda.enums.PollType`, *optional*):
                Poll type, :obj:`~Badmunda.enums.PollType.QUIZ` or :obj:`~Badmunda.enums.PollType.REGULAR`.
                Defaults to :obj:`~Badmunda.enums.PollType.REGULAR`.

            allows_multiple_answers (``bool``, *optional*):
                True, if the poll allows multiple answers, ignored for polls in quiz mode.
                Defaults to False.

            correct_option_id (``int``, *optional*):
                0-based identifier of the correct answer option, required for polls in quiz mode.

            question_parse_mode (:obj:`~Badmunda.enums.ParseMode`, *optional*):
                By default, texts are parsed using both Markdown and HTML styles.
                You can combine both syntaxes together.

            question_entities (List of :obj:`~Badmunda.types.MessageEntity`):
                List of special entities that appear in the poll question, which can be specified instead of
                *parse_mode*.

            explanation (``str``, *optional*):
                Text that is shown when a user chooses an incorrect answer or taps on the lamp icon in a quiz-style
                poll, 0-200 characters with at most 2 line feeds after entities parsing.

            explanation_parse_mode (:obj:`~Badmunda.enums.ParseMode`, *optional*):
                By default, texts are parsed using both Markdown and HTML styles.
                You can combine both syntaxes together.

            explanation_entities (List of :obj:`~Badmunda.types.MessageEntity`):
                List of special entities that appear in the poll explanation, which can be specified instead of
                *parse_mode*.

            open_period (``int``, *optional*):
                Amount of time in seconds the poll will be active after creation, 5-600.
                Can't be used together with *close_date*.

            close_date (:py:obj:`~datetime.datetime`, *optional*):
                Point in time when the poll will be automatically closed.
                Must be at least 5 and no more than 600 seconds in the future.
                Can't be used together with *open_period*.

            is_closed (``bool``, *optional*):
                Pass True, if the poll needs to be immediately closed.
                This can be useful for poll preview.

            disable_notification (``bool``, *optional*):
                Sends the message silently.
                Users will receive a notification with no sound.

            protect_content (``bool``, *optional*):
                Protects the contents of the sent message from forwarding and saving.

            message_thread_id (``int``, *optional*):
                Unique identifier for the target message thread (topic) of the forum.
                For supergroups only.

            effect_id (``int``, *optional*):
                Unique identifier of the message effect.
                For private chats only.

            reply_parameters (:obj:`~Badmunda.types.ReplyParameters`, *optional*):
                Describes reply parameters for the message that is being sent.

            schedule_date (:py:obj:`~datetime.datetime`, *optional*):
                Date when the message will be automatically sent.

            business_connection_id (``str``, *optional*):
                Unique identifier of the business connection on behalf of which the message will be sent.

            options_parse_mode (:obj:`~Badmunda.enums.ParseMode`, *optional*):
                By default, texts are parsed using both Markdown and HTML styles.
                You can combine both syntaxes together.

            allow_paid_broadcast (``bool``, *optional*):
                If True, you will be allowed to send up to 1000 messages per second.
                Ignoring broadcasting limits for a fee of 0.1 Telegram Stars per message.
                The relevant Stars will be withdrawn from the bot's balance.
                For bots only.

            paid_message_star_count (``int``, *optional*):
                The number of Telegram Stars the user agreed to pay to send the messages.

            reply_markup (:obj:`~Badmunda.types.InlineKeyboardMarkup` | :obj:`~Badmunda.types.ReplyKeyboardMarkup` | :obj:`~Badmunda.types.ReplyKeyboardRemove` | :obj:`~Badmunda.types.ForceReply`, *optional*):
                Additional interface options. An object for an inline keyboard, custom reply keyboard,
                instructions to remove reply keyboard or to force a reply from the user.

        Returns:
            :obj:`~Badmunda.types.Message`: On success, the sent poll message is returned.

        Example:
            .. code-block:: python

                await app.send_poll(chat_id, "Is this a poll question?", ["Yes", "No", "Maybe"])
        """
        if any(
            (
                reply_to_message_id is not None,
                reply_to_chat_id is not None,
                quote_text is not None,
                quote_parse_mode is not None,
                quote_entities is not None,
                quote_offset is not None,
            )
        ):
            if reply_to_message_id is not None:
                log.warning(
                    "`reply_to_message_id` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if reply_to_chat_id is not None:
                log.warning(
                    "`reply_to_chat_id` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if quote_text is not None:
                log.warning(
                    "`quote_text` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if quote_parse_mode is not None:
                log.warning(
                    "`quote_parse_mode` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if quote_entities is not None:
                log.warning(
                    "`quote_entities` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if quote_offset is not None:
                log.warning(
                    "`quote_offset` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            reply_parameters = types.ReplyParameters(
                message_id=reply_to_message_id,
                chat_id=reply_to_chat_id,
                quote=quote_text,
                quote_parse_mode=quote_parse_mode,
                quote_entities=quote_entities,
                quote_position=quote_offset
            )

        question, question_entities = (await utils.parse_text_entities(
            self, question, question_parse_mode, question_entities
        )).values()

        solution, solution_entities = (await utils.parse_text_entities(
            self, explanation, explanation_parse_mode, explanation_entities
        )).values()

        answers = []

        for i, opt in enumerate(options):
            option, option_entities = (await utils.parse_text_entities(
                self, opt, options_parse_mode, None
            )).values()

            answers.append(
                raw.types.PollAnswer(
                    text=raw.types.TextWithEntities(text=option, entities=option_entities or []),
                    option=bytes([i]),
                )
            )

        r = await self.invoke(
            raw.functions.messages.SendMedia(
                peer=await self.resolve_peer(chat_id),
                media=raw.types.InputMediaPoll(
                    poll=raw.types.Poll(
                        id=self.rnd_id(),
                        question=raw.types.TextWithEntities(
                            text=question,
                            entities=question_entities or []
                        ),
                        answers=answers,
                        closed=is_closed,
                        public_voters=not is_anonymous,
                        multiple_choice=allows_multiple_answers,
                        quiz=type == enums.PollType.QUIZ or False,
                        close_period=open_period,
                        close_date=utils.datetime_to_timestamp(close_date)
                    ),
                    correct_answers=[bytes([correct_option_id])] if correct_option_id is not None else None,
                    solution=solution,
                    solution_entities=solution_entities or []
                ),
                message="",
                silent=disable_notification,
                reply_to=await utils.get_reply_to(
                    self,
                    reply_parameters,
                    message_thread_id
                ),
                random_id=self.rnd_id(),
                schedule_date=utils.datetime_to_timestamp(schedule_date),
                noforwards=protect_content,
                allow_paid_floodskip=allow_paid_broadcast,
                allow_paid_stars=paid_message_star_count,
                reply_markup=await reply_markup.write(self) if reply_markup else None,
                effect=effect_id
            ),
            business_connection_id=business_connection_id
        )

        messages = await utils.parse_messages(client=self, messages=r)

        return messages[0] if messages else None



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import re
from datetime import datetime
from typing import BinaryIO, Callable, List, Optional, Union

import Badmunda
from Badmunda import StopTransmission, enums, raw, types, utils
from Badmunda.errors import FilePartMissing
from Badmunda.file_id import FileType

log = logging.getLogger(__name__)

class SendVoice:
    async def send_voice(
        self: "Badmunda.Client",
        chat_id: Union[int, str],
        voice: Union[str, BinaryIO],
        caption: str = "",
        parse_mode: Optional["enums.ParseMode"] = None,
        caption_entities: List["types.MessageEntity"] = None,
        duration: int = 0,
        disable_notification: bool = None,
        message_thread_id: int = None,
        direct_messages_chat_topic_id: int = None,
        effect_id: int = None,
        reply_parameters: "types.ReplyParameters" = None,
        schedule_date: datetime = None,
        protect_content: bool = None,
        view_once: bool = None,
        business_connection_id: str = None,
        allow_paid_broadcast: bool = None,
        paid_message_star_count: int = None,
        reply_markup: Union[
            "types.InlineKeyboardMarkup",
            "types.ReplyKeyboardMarkup",
            "types.ReplyKeyboardRemove",
            "types.ForceReply"
        ] = None,
        progress: Callable = None,
        progress_args: tuple = (),

        reply_to_message_id: int = None,
        reply_to_chat_id: Union[int, str] = None,
        reply_to_story_id: int = None,
        quote_text: str = None,
        quote_entities: List["types.MessageEntity"] = None,
        quote_offset: int = None,
    ) -> Optional["types.Message"]:
        """Send audio files.

        .. include:: /_includes/usable-by/users-bots.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".
                For a contact that exists in your Telegram address book you can use his phone number (str).

            voice (``str`` | ``BinaryIO``):
                Audio file to send.
                Pass a file_id as string to send an audio that exists on the Telegram servers,
                pass an HTTP URL as a string for Telegram to get an audio from the Internet,
                pass a file path as string to upload a new audio that exists on your local machine, or
                pass a binary file-like object with its attribute ".name" set for in-memory uploads.

            caption (``str``, *optional*):
                Voice message caption, 0-1024 characters.

            parse_mode (:obj:`~Badmunda.enums.ParseMode`, *optional*):
                By default, texts are parsed using both Markdown and HTML styles.
                You can combine both syntaxes together.

            caption_entities (List of :obj:`~Badmunda.types.MessageEntity`):
                List of special entities that appear in the caption, which can be specified instead of *parse_mode*.

            duration (``int``, *optional*):
                Duration of the voice message in seconds.

            disable_notification (``bool``, *optional*):
                Sends the message silently.
                Users will receive a notification with no sound.

            message_thread_id (``int``, *optional*):
                Unique identifier for the target message thread (topic) of the forum.
                For forums only.

            direct_messages_chat_topic_id (``int``, *optional*):
                Unique identifier of the topic in a channel direct messages chat administered by the current user.
                For directs only only.

            effect_id (``int``, *optional*):
                Unique identifier of the message effect.
                For private chats only.

            reply_parameters (:obj:`~Badmunda.types.ReplyParameters`, *optional*):
                Describes reply parameters for the message that is being sent.

            schedule_date (:py:obj:`~datetime.datetime`, *optional*):
                Date when the message will be automatically sent.

            protect_content (``bool``, *optional*):
                Protects the contents of the sent message from forwarding and saving.

            view_once (``bool``, *optional*):
                Self-Destruct Timer.
                If True, the voice note will self-destruct after it was listened.

            business_connection_id (``str``, *optional*):
                Unique identifier of the business connection on behalf of which the message will be sent.

            allow_paid_broadcast (``bool``, *optional*):
                If True, you will be allowed to send up to 1000 messages per second.
                Ignoring broadcasting limits for a fee of 0.1 Telegram Stars per message.
                The relevant Stars will be withdrawn from the bot's balance.
                For bots only.

            paid_message_star_count (``int``, *optional*):
                The number of Telegram Stars the user agreed to pay to send the messages.

            reply_markup (:obj:`~Badmunda.types.InlineKeyboardMarkup` | :obj:`~Badmunda.types.ReplyKeyboardMarkup` | :obj:`~Badmunda.types.ReplyKeyboardRemove` | :obj:`~Badmunda.types.ForceReply`, *optional*):
                Additional interface options. An object for an inline keyboard, custom reply keyboard,
                instructions to remove reply keyboard or to force a reply from the user.

            progress (``Callable``, *optional*):
                Pass a callback function to view the file transmission progress.
                The function must take *(current, total)* as positional arguments (look at Other Parameters below for a
                detailed description) and will be called back each time a new file chunk has been successfully
                transmitted.

            progress_args (``tuple``, *optional*):
                Extra custom arguments for the progress callback function.
                You can pass anything you need to be available in the progress callback scope; for example, a Message
                object or a Client instance in order to edit the message with the updated progress status.

        Other Parameters:
            current (``int``):
                The amount of bytes transmitted so far.

            total (``int``):
                The total size of the file.

            *args (``tuple``, *optional*):
                Extra custom arguments as defined in the ``progress_args`` parameter.
                You can either keep ``*args`` or add every single extra argument in your function signature.

        Returns:
            :obj:`~Badmunda.types.Message` | ``None``: On success, the sent voice message is returned, otherwise, in
            case the upload is deliberately stopped with :meth:`~Badmunda.Client.stop_transmission`, None is returned.

        Example:
            .. code-block:: python

                # Send voice note by uploading from local file
                await app.send_voice("me", "voice.ogg")

                # Add caption to the voice note
                await app.send_voice("me", "voice.ogg", caption="voice caption")

                # Set voice note duration
                await app.send_voice("me", "voice.ogg", duration=20)

                # Send self-destructing voice note
                await app.send_voice("me", "voice.ogg", view_once=True)
        """
        if any(
            (
                reply_to_message_id is not None,
                reply_to_chat_id is not None,
                reply_to_story_id is not None,
                quote_text is not None,
                quote_entities is not None,
                quote_offset is not None,
            )
        ):
            if reply_to_message_id is not None:
                log.warning(
                    "`reply_to_message_id` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if reply_to_chat_id is not None:
                log.warning(
                    "`reply_to_chat_id` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if reply_to_story_id is not None:
                log.warning(
                    "`reply_to_story_id` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if quote_text is not None:
                log.warning(
                    "`quote_text` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if quote_entities is not None:
                log.warning(
                    "`quote_entities` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            if quote_offset is not None:
                log.warning(
                    "`quote_offset` is deprecated and will be removed in future updates. Use `reply_parameters` instead."
                )

            reply_parameters = types.ReplyParameters(
                message_id=reply_to_message_id,
                chat_id=reply_to_chat_id,
                story_id=reply_to_story_id,
                quote=quote_text,
                quote_parse_mode=parse_mode,
                quote_entities=quote_entities,
                quote_position=quote_offset
            )

        file = None

        try:
            if isinstance(voice, str):
                if os.path.isfile(voice):
                    mime_type = self.guess_mime_type(voice) or "audio/ogg"
                    if mime_type == "audio/mpeg":
                        mime_type = "audio/ogg"
                    file = await self.save_file(voice, progress=progress, progress_args=progress_args)
                    media = raw.types.InputMediaUploadedDocument(
                        mime_type=mime_type,
                        file=file,
                        attributes=[
                            raw.types.DocumentAttributeAudio(
                                voice=True,
                                duration=duration
                            )
                        ],
                        ttl_seconds=(1 << 31) - 1 if view_once else None
                    )
                elif re.match("^https?://", voice):
                    media = raw.types.InputMediaDocumentExternal(
                        url=voice
                    )
                else:
                    media = utils.get_input_media_from_file_id(voice, FileType.VOICE)
            else:
                mime_type = self.guess_mime_type(voice.name) or "audio/ogg"
                if mime_type == "audio/mpeg":
                    mime_type = "audio/ogg"
                file = await self.save_file(voice, progress=progress, progress_args=progress_args)
                media = raw.types.InputMediaUploadedDocument(
                    mime_type=mime_type,
                    file=file,
                    attributes=[
                        raw.types.DocumentAttributeAudio(
                            voice=True,
                            duration=duration
                        )
                    ],
                    ttl_seconds=(1 << 31) - 1 if view_once else None
                )

            while True:
                try:
                    peer = await self.resolve_peer(chat_id)
                    r = await self.invoke(
                        raw.functions.messages.SendMedia(
                            peer=peer,
                            media=media,
                            silent=disable_notification or None,
                            reply_to=await utils.get_reply_to(
                                self,
                                reply_parameters,
                                message_thread_id,
                                direct_messages_chat_topic_id
                            ),
                            random_id=self.rnd_id(),
                            schedule_date=utils.datetime_to_timestamp(schedule_date),
                            noforwards=protect_content,
                            allow_paid_floodskip=allow_paid_broadcast,
                            allow_paid_stars=paid_message_star_count,
                            reply_markup=await reply_markup.write(self) if reply_markup else None,
                            effect=effect_id,
                            **await utils.parse_text_entities(self, caption, parse_mode, caption_entities)
                        ),
                        business_connection_id=business_connection_id
                    )
                except FilePartMissing as e:
                    await self.save_file(voice, file_id=file.id, file_part=e.value)
                else:
                    messages = await utils.parse_messages(client=self, messages=r)

                    return messages[0] if messages else None
        except StopTransmission:
            return None



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import random
from typing import Dict, Optional

import Badmunda
from Badmunda import raw, types, utils
from Badmunda.errors import MessageIdsEmpty, ChannelPrivate

from ..object import Object


class GiveawayPrizeStars(Object):
    """A Telegram Stars were received by the current user from a giveaway.

    Parameters:
        star_count (``int``):
            Number of Telegram Stars that were gifted.

        transaction_id (``str``):
            Identifier of the transaction for Telegram Stars purchase.
            For receiver only

        boosted_chat (:obj:`~Badmunda.types.Chat`):
            Supergroup or channel chat, which was automatically boosted by the winners of the giveaway.

        giveaway_message_id (``int``):
            Identifier of the message with the giveaway in the boosted chat.

        giveaway_message (:obj:`~Badmunda.types.Message`, *optional*):
            Message with the giveaway in the boosted chat.

        is_unclaimed (``bool``, *optional*):
            True, if the corresponding winner wasn't chosen and the Telegram Stars were received by the owner of the boosted chat.

        sticker (:obj:`~Badmunda.types.Sticker`):
            A sticker to be shown in the message.
    """
    def __init__(
        self,
        *,
        star_count: int,
        transaction_id: str,
        boosted_chat: "types.Chat",
        giveaway_message_id: int,
        giveaway_message: Optional["types.Message"] = None,
        is_unclaimed: Optional[bool] = None,
        sticker: Optional["types.Sticker"] = None
    ):
        super().__init__()

        self.star_count = star_count
        self.transaction_id = transaction_id
        self.boosted_chat = boosted_chat
        self.giveaway_message_id = giveaway_message_id
        self.giveaway_message = giveaway_message
        self.is_unclaimed = is_unclaimed
        self.sticker = sticker

    @staticmethod
    async def _parse(
        client: "Badmunda.Client",
        action: "raw.types.MessageActionPrizeStars",
        chats: Dict[int, "raw.base.Chat"],
    ) -> "GiveawayPrizeStars":
        raw_stickers = await client.invoke(
            raw.functions.messages.GetStickerSet(
                stickerset=raw.types.InputStickerSetPremiumGifts(),
                hash=0
            )
        )

        parsed_message = None

        try:
            parsed_message = await client.get_messages(
                chat_id=utils.get_peer_id(action.boost_peer),
                message_ids=action.giveaway_msg_id,
                replies=0
            )
        except (MessageIdsEmpty, ChannelPrivate):
            pass

        return GiveawayPrizeStars(
            star_count=action.stars,
            transaction_id=action.transaction_id,
            boosted_chat=types.Chat._parse_chat(client, chats.get(utils.get_raw_peer_id(action.boost_peer))),
            giveaway_message_id=action.message_id,
            giveaway_message=parsed_message,
            sticker=random.choice(
                types.List(
                    [
                        await types.Sticker._parse(
                            client,
                            doc,
                            {
                                type(i): i for i in doc.attributes
                            }
                        ) for doc in raw_stickers.documents
                    ]
                )
            )
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Union

import Badmunda
from Badmunda import enums, raw


class SendPaidReaction:
    async def send_paid_reaction(
        self: "Badmunda.Client",
        chat_id: Union[int, str],
        message_id: int,
        amount: int,
        privacy: "enums.PaidReactionPrivacy" = None,
        send_as: Optional[Union[int, str]] = None,
    ) -> bool:
        """Send a paid reaction to a message.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.

            message_id (``int``):
                Identifier of the message.

            amount (``int``):
                Amount of stars to send.

            privacy (:obj:`~Badmunda.enums.PaidReactionPrivacy`, *optional*):
                Reaction privacy type.

            send_as (``int`` | ``str``, *optional*):
                Unique identifier (int) or username (str) of the send_as chat.
                Applicable when privacy is :obj:`~Badmunda.enums.PaidReactionPrivacy.CHAT`.

        Returns:
            ``bool``: On success, True is returned.

        Example:
            .. code-block:: python

                # Send paid reaction with 1 star
                await app.send_paid_reaction(chat_id, message_id, amount=1)
        """
        if privacy:
            is_queryable = privacy in [enums.PaidReactionPrivacy.CHAT]

            privacy = privacy.value(peer=await self.resolve_peer(send_as)) if is_queryable else privacy.value()

        rpc = raw.functions.messages.SendPaidReaction(
            peer=await self.resolve_peer(chat_id),
            msg_id=message_id,
            count=amount,
            random_id=self.rnd_id(),
            private=privacy
        )

        await self.invoke(rpc)

        return True



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from typing import Dict, List, Optional

import Badmunda
from Badmunda import raw, types, utils

from ..object import Object


class ChecklistTask(Object):
    """Describes a task in a checklist.

    Parameters:
        id (``int``):
            Unique identifier of the task.

        text (``str``):
            Text of the task.

        entities (List of :obj:`~Badmunda.types.MessageEntity`, *optional*):
            Entities in the text of the task.
            May contain only Bold, Italic, Underline, Strikethrough, Spoiler, CustomEmoji, Url, EmailAddress, Mention, Hashtag, Cashtag and PhoneNumber entities.

        completed_by_user (:obj:`~Badmunda.types.User`, *optional*):
            The user that completed the task.
            None if the task isn't completed.

        completion_date (:py:obj:`~datetime.datetime`, *optional*):
            Date when the task was completed.
            None if the task isn't completed.
    """

    def __init__(
        self,
        *,
        id: int,
        text: str,
        entities: Optional[List["types.MessageEntity"]] = None,
        completed_by_user: Optional["types.User"] = None,
        completion_date: Optional[datetime] = None,
    ):
        super().__init__()

        self.id = id
        self.text = text
        self.entities = entities
        self.completed_by_user = completed_by_user
        self.completion_date = completion_date

    @staticmethod
    def _parse(
        client: "Badmunda.Client",
        item: "raw.types.TodoItem",
        completion: "raw.types.TodoCompletion",
        users: Dict[int, "raw.base.User"],
    ) -> "ChecklistTask":
        raw.types.TodoItem
        raw.types.TodoCompletion

        text, entities = (
            utils.parse_text_with_entities(client, item.title, users)
        ).values()

        return ChecklistTask(
            id=item.id,
            text=text,
            entities=entities,
            completed_by_user=types.User._parse(client, users.get(getattr(completion, "completed_by", None))),
            completion_date=utils.timestamp_to_datetime(getattr(completion, "date", None))
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from typing import Optional

import Badmunda
from Badmunda import raw, utils

from ..object import Object


class EmojiStatus(Object):
    """A user emoji status.

    Parameters:
        custom_emoji_id (``int``, *optional*):
            Custom emoji id.

        until_date (:py:obj:`~datetime.datetime`, *optional*):
            Valid until date.

        title (``str``, *optional*):
            Title of the collectible.

        gift_id (``int``, *optional*):
            Gift collectible id.

        name (``str``, *optional*):
            Name of the collectible.

        pattern_custom_emoji_id (``int``, *optional*):
            Pattern emoji id.

        center_color (``int``, *optional*):
            Center color of the collectible emoji in decimal format.

        edge_color (``int``, *optional*):
            Edge color of the collectible emoji in decimal format.

        pattern_color (``int``, *optional*):
            Pattern color of the collectible emoji in decimal format.

        text_color (``int``, *optional*):
            Text color of the collectible emoji in decimal format.
    """

    def __init__(
        self,
        *,
        client: "Badmunda.Client" = None,
        custom_emoji_id: Optional[int] = None,
        gift_id: Optional[int] = None,
        until_date: Optional[datetime] = None,
        title: Optional[str] = None,
        name: Optional[str] = None,
        pattern_custom_emoji_id: Optional[int] = None,
        center_color: Optional[int] = None,
        edge_color: Optional[int] = None,
        pattern_color: Optional[int] = None,
        text_color: Optional[int] = None
    ):
        super().__init__(client)

        self.custom_emoji_id = custom_emoji_id
        self.gift_id = gift_id
        self.until_date = until_date
        self.title = title
        self.name = name
        self.pattern_custom_emoji_id = pattern_custom_emoji_id
        self.center_color = center_color
        self.edge_color = edge_color
        self.pattern_color = pattern_color
        self.text_color = text_color

    @staticmethod
    def _parse(client, emoji_status: "raw.base.EmojiStatus") -> Optional["EmojiStatus"]:
        if isinstance(emoji_status, raw.types.EmojiStatus):
            return EmojiStatus(
                client=client,
                custom_emoji_id=emoji_status.document_id,
                until_date=utils.timestamp_to_datetime(getattr(emoji_status, "until", None))
            )

        if isinstance(emoji_status, raw.types.EmojiStatusCollectible):
            return EmojiStatus(
                client=client,
                custom_emoji_id=emoji_status.document_id,
                gift_id=emoji_status.collectible_id,
                until_date=utils.timestamp_to_datetime(getattr(emoji_status, "until", None)),
                title=emoji_status.title,
                name=emoji_status.slug,
                pattern_custom_emoji_id=emoji_status.pattern_document_id,
                center_color=emoji_status.center_color,
                edge_color=emoji_status.edge_color,
                pattern_color=emoji_status.pattern_color,
                text_color=emoji_status.text_color
            )

        return None

    def write(self):
        if self.gift_id:
            return raw.types.InputEmojiStatusCollectible(
                collectible_id=self.gift_id,
                until=utils.datetime_to_timestamp(self.until_date)
            )

        return raw.types.EmojiStatus(
            document_id=self.custom_emoji_id,
            until=utils.datetime_to_timestamp(self.until_date)
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import Badmunda
from Badmunda import raw, types


class GetPaymentForm:
    async def get_payment_form(
        self: "Badmunda.Client",
        input_invoice: "types.InputInvoice"
    ) -> "types.PaymentForm":
        """Get an invoice payment form.

        This method must be called when the user presses inline button of the type InlineKeyboardButton with buy parameter,
        or wants to buy access to media in a paid media message.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            input_invoice (:obj:`~Badmunda.types.InputInvoice`):
                The invoice.

        Returns:
            :obj:`~Badmunda.types.PaymentForm`: On success, a payment form is returned.

        Example:
            .. code-block:: python

                # Get payment form from message
                await app.get_payment_form(
                    types.InputInvoiceMessage(
                        chat_id=chat_id,
                        message_id=123
                    )
                )

                # Get payment form from link
                await app.get_payment_form(
                    types.InputInvoiceName(
                        name="https://t.me/$xvbzUtt5sUlJCAAATqZrWRy9Yzk"
                    )
                )
        """
        r = await self.invoke(
            raw.functions.payments.GetPaymentForm(
                invoice=await input_invoice.write(self)
            )
        )

        return types.PaymentForm._parse(self, r)



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Union

import Badmunda
from Badmunda import enums, raw, types
from Badmunda.errors import ChannelInvalid, ChannelPrivate

from ..object import Object


class MediaArea(Object):
    """Describes the media area added to a story.

    Parameters:
        x (``float``):
            The abscissa of the rectangle's center, as a percentage of the media width (0-100).

        y (``float``):
            The ordinate of the rectangle's center, as a percentage of the media height (0-100).

        width (``float``):
            The width of the rectangle, as a percentage of the media width (0-100).

        height (``float``):
            The height of the rectangle, as a percentage of the media height (0-100).

        rotation (``float``):
            Clockwise rotation angle of the rectangle, in degrees (0-360).

        type (:obj:`~Badmunda.enums.MediaAreaType`):
            The type of the media area.

        radius (``float``, *optional*):
            The radius of the rectangle corner rounding, as a percentage of the media width.

        sender_chat (:obj:`~Badmunda.types.Chat`, *optional*):
            The channel that posted the message.
            For post type only.

        message_id (``int``, *optional*):
            Identifier of the message in channel.
            For post type only.

        message (:obj:`~Badmunda.types.Message`, *optional*):
            Message attached to the story.
            Can be empty if the message was deleted or unavailable (channel is private).
            For post type only.

        location (:obj:`~Badmunda.types.Location`, *optional*):
            Represents a location tag attached to a story.
            For location type only.

        reaction (:obj:`~Badmunda.types.Reaction`, *optional*):
            The reaction that should be sent when this area is clicked.
            For reaction type only.

        is_dark (``bool``, *optional*):
            True if the reaction bubble has a dark background.
            For reaction type only.

        is_flipped (``bool``, *optional*):
            True if the reaction bubble is mirrored.
            For reaction type only.

        url (``str``, *optional*):
            URL to open when clicked.
            For url type only.

        emoji (``str``, *optional*):
            Weather emoji, should be rendered as an animated emoji.
            For weather type only.

        temperature (``float``, *optional*):
            Temperature in degrees Celsius.
            For weather type only.

        color (``int``, *optional*):
            The argb color in decimal format.
            For weather type only.

        gift (:obj:`~Badmunda.types.Gift`, *optional*):
            Information about this gift.
    """

    def __init__(
        self,
        *,
        client: "Badmunda.Client" = None,
        x: float,
        y: float,
        width: float,
        height: float,
        rotation: float,
        type: "enums.MediaAreaType",
        radius: float = None,
        sender_chat: Optional["types.Chat"] = None,
        message_id: Optional[int] = None,
        message: Optional["types.Message"] = None,
        location: Optional["types.Location"] = None,
        reaction: Optional["types.Reaction"] = None,
        is_dark: Optional[bool] = None,
        is_flipped: Optional[bool] = None,
        url: Optional[str] = None,
        venue: Optional["types.Venue"] = None,
        emoji: Optional[str] = None,
        temperature: Optional[float] = None,
        color: Optional[int] = None,
        gift: Optional["types.Gift"] = None
    ):
        super().__init__(client)

        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.rotation = rotation
        self.type = type
        self.radius = radius
        self.sender_chat = sender_chat
        self.message_id = message_id
        self.message = message
        self.location = location
        self.reaction = reaction
        self.is_dark = is_dark
        self.is_flipped = is_flipped
        self.url = url
        self.venue = venue
        self.emoji = emoji
        self.temperature = temperature
        self.color = color
        self.gift = gift

    @staticmethod
    async def _parse(
        client: "Badmunda.Client",
        area: "raw.base.MediaArea",
        chats: dict = None
    ) -> "MediaArea":
        sender_chat = None
        message_id = None
        message = None
        location = None
        reaction = None
        is_dark = None
        is_flipped = None
        url = None
        venue = None
        emoji = None
        temperature = None
        color = None
        gift = None

        if isinstance(area, raw.types.MediaAreaChannelPost):
            sender_chat = types.Chat._parse_channel_chat(client, chats.get(area.channel_id))
            message_id = area.msg_id

            try:
                message = await client.get_messages(chat_id=sender_chat.id, message_ids=message_id)
            except (ChannelPrivate, ChannelInvalid):
                pass
        elif isinstance(area, raw.types.MediaAreaGeoPoint):
            location = types.Location._parse(client, area.geo)
        elif isinstance(area, raw.types.MediaAreaSuggestedReaction):
            reaction = types.Reaction._parse(client, area.reaction)
            is_dark = getattr(area, "dark", None)
            is_flipped = getattr(area, "flipped", None)
        elif isinstance(area, raw.types.MediaAreaUrl):
            url = area.url
        elif isinstance(area, raw.types.MediaAreaVenue):
            venue = types.Venue._parse(client, area)
        elif isinstance(area, raw.types.MediaAreaWeather):
            emoji = area.emoji
            temperature = area.temperature_c
            color = area.color
        elif isinstance(area, raw.types.MediaAreaStarGift):
            gift = await client.get_upgraded_gift(area.slug)

        return MediaArea(
            x=area.coordinates.x,
            y=area.coordinates.y,
            width=area.coordinates.w,
            height=area.coordinates.h,
            rotation=area.coordinates.rotation,
            type=enums.MediaAreaType(type(area)),
            radius=getattr(area.coordinates, "radius", None),
            sender_chat=sender_chat,
            message_id=message_id,
            message=message,
            location=location,
            reaction=reaction,
            is_dark=is_dark,
            is_flipped=is_flipped,
            url=url,
            venue=venue,
            emoji=emoji,
            temperature=temperature,
            color=color,
            gift=gift,
            client=client
        )

    async def write(
        self, client: "Badmunda.Client"
    ) -> Optional[
        Union[
            "raw.types.InputMediaAreaChannelPost",
            "raw.types.MediaAreaGeoPoint",
            "raw.types.MediaAreaSuggestedReaction",
            "raw.types.MediaAreaUrl",
            "raw.types.MediaAreaWeather",
            "raw.types.MediaAreaStarGift"
        ]
    ]:
        coordinates = raw.types.MediaAreaCoordinates(
            x=self.x,
            y=self.y,
            w=self.width,
            h=self.height,
            rotation=self.rotation,
            radius=self.radius
        )

        if self.type == enums.MediaAreaType.POST:
            return raw.types.InputMediaAreaChannelPost(
                coordinates=coordinates,
                channel=await client.resolve_peer(self.sender_chat.id),
                msg_id=self.message_id
            )
        elif self.type == enums.MediaAreaType.LOCATION:
            return raw.types.MediaAreaGeoPoint(
                coordinates=coordinates,
                geo=raw.types.InputGeoPoint(
                    lat=self.location.latitude,
                    long=self.location.longitude,
                    accuracy_radius=self.location.accuracy_radius
                )
            )
        elif self.type == enums.MediaAreaType.REACTION:
            if self.reaction.custom_emoji_id:
                reaction = raw.types.ReactionCustomEmoji(
                    document_id=self.reaction.custom_emoji_id
                )
            else:
                reaction = raw.types.ReactionEmoji(
                    emoticon=self.reaction.emoji
                )

            return raw.types.MediaAreaSuggestedReaction(
                coordinates=coordinates,
                reaction=reaction,
                dark=self.is_dark,
                flipped=self.is_flipped
            )
        elif self.type == enums.MediaAreaType.URL:
            return raw.types.MediaAreaUrl(
                coordinates=coordinates,
                url=self.url
            )
        elif self.type == enums.MediaAreaType.WEATHER:
            return raw.types.MediaAreaWeather(
                coordinates=coordinates,
                emoji=self.emoji,
                temperature_c=self.temperature,
                color=self.color
            )
        elif self.type == enums.MediaAreaType.GIFT:
            return raw.types.MediaAreaStarGift(
                coordinates=coordinates,
                slug=self.gift.name
            )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import Badmunda

from Badmunda import raw, types, errors
from ..object import Object



class GiveawayCompleted(Object):
    """This object represents a service message about the completion of a giveaway without public winners.

    Parameters:
        winner_count (``int``):
            Number of winners in the giveaway.

        unclaimed_prize_count (``int``, *optional*):
            Number of undistributed prizes.

        giveaway_message_id (``int``, *optional*):
            Identifier of the message with the giveaway in the chat.

        giveaway_message (:obj:`~Badmunda.types.Message`, *optional*):
            Message with the giveaway that was completed, if it wasn't deleted.

        is_star_giveaway (``bool``, *optional*):
            True, if the giveaway is a Telegram Star giveaway. Otherwise, currently, the giveaway is a Telegram Premium giveaway.

    """

    def __init__(
        self,
        *,
        client: "Badmunda.Client" = None,
        winner_count: int,
        unclaimed_prize_count: int = None,
        giveaway_message_id: int = None,
        giveaway_message: "types.Message" = None,
        is_star_giveaway: bool = None
    ):
        super().__init__(client)

        self.winner_count = winner_count
        self.unclaimed_prize_count = unclaimed_prize_count
        self.giveaway_message_id = giveaway_message_id
        self.giveaway_message = giveaway_message
        self.is_star_giveaway = is_star_giveaway


    @staticmethod
    async def _parse(
        client,
        giveaway_results: "raw.types.MessageActionGiveawayResults",
        chat: "types.Chat" = None,
        message_id: int = None
    ) -> "GiveawayCompleted":
        if not isinstance(giveaway_results, raw.types.MessageActionGiveawayResults):
            return

        giveaway_message = None

        if chat and message_id:
            try:
                giveaway_message = await client.get_messages(
                    chat_id=chat.id,
                    message_ids=message_id,
                    replies=0
                )
            except (errors.ChannelPrivate, errors.ChannelInvalid):
                pass

        return GiveawayCompleted(
            winner_count=giveaway_results.winners_count,
            unclaimed_prize_count=giveaway_results.unclaimed_count,
            giveaway_message_id=message_id,
            giveaway_message=giveaway_message,
            is_star_giveaway=getattr(giveaway_results, "stars", None),
            client=client,
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import AsyncGenerator, Optional, Union

import Badmunda
from Badmunda import raw, types, utils


class GetDirectMessagesTopics:
    async def get_direct_messages_topics(
        self: "Badmunda.Client",
        chat_id: Union[int, str],
        limit: int = 0,
        exclude_pinned: Optional[bool] = None
    ) -> AsyncGenerator["types.DirectMessagesTopic", None]:
        """Get one or more topic from a direct messages channel chat.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.

            limit (``int``, *optional*):
                Limits the number of topics to be retrieved.
                By default, no limit is applied and all topics are returned.

            exclude_pinned (``bool``, *optional*):
                Exclude pinned topics.

        Returns:
            ``Generator``: A generator yielding :obj:`~Badmunda.types.DirectMessagesTopic` objects.

        Example:
            .. code-block:: python

                # Iterate through all topics
                async for topic in app.get_direct_messages_topics(chat_id):
                    print(topic)
        """
        current = 0
        total = limit or (1 << 31) - 1
        limit = min(100, total)

        offset_date = 0
        offset_id = 0
        offset_peer = raw.types.InputPeerEmpty()

        while True:
            r = await self.invoke(
                raw.functions.messages.GetSavedDialogs(
                    offset_date=offset_date,
                    offset_id=offset_id,
                    offset_peer=offset_peer,
                    limit=limit,
                    hash=0,
                    exclude_pinned=exclude_pinned,
                    parent_peer=await self.resolve_peer(chat_id)
                )
            )

            users = {i.id: i for i in r.users}
            chats = {i.id: i for i in r.chats}

            messages = {}

            for message in r.messages:
                if isinstance(message, raw.types.MessageEmpty):
                    continue

                messages[message.id] = await types.Message._parse(self, message, users, chats)

            topics = []

            for topic in r.dialogs:
                topics.append(types.DirectMessagesTopic._parse(topic, messages, users, chats))

            if not topics:
                return

            last = topics[-1]

            offset_id = last.last_message.id
            offset_date = utils.datetime_to_timestamp(last.last_message.date)
            offset_peer = await self.resolve_peer(last.id)

            for topic in topics:
                yield topic

                current += 1

                if current >= total:
                    return



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union, List

import Badmunda
from Badmunda import raw
from Badmunda import types


class GetGameHighScores:
    async def get_game_high_scores(
        self: "Badmunda.Client",
        user_id: Union[int, str],
        chat_id: Union[int, str],
        message_id: int = None
    ) -> List["types.GameHighScore"]:
        """Get data for high score tables.

        .. include:: /_includes/usable-by/bots.rst

        Parameters:
            user_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".
                For a contact that exists in your Telegram address book you can use his phone number (str).

            chat_id (``int`` | ``str``, *optional*):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".
                For a contact that exists in your Telegram address book you can use his phone number (str).
                Required if inline_message_id is not specified.

            message_id (``int``, *optional*):
                Identifier of the sent message.
                Required if inline_message_id is not specified.

        Returns:
            List of :obj:`~Badmunda.types.GameHighScore`: On success.

        Example:
            .. code-block:: python

                scores = await app.get_game_high_scores(user_id, chat_id, message_id)
                print(scores)
        """
        # TODO: inline_message_id

        r = await self.invoke(
            raw.functions.messages.GetGameHighScores(
                peer=await self.resolve_peer(chat_id),
                id=message_id,
                user_id=await self.resolve_peer(user_id)
            )
        )

        return types.List(types.GameHighScore._parse(self, score, r.users) for score in r.scores)



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

import Badmunda
from Badmunda import types, raw
from ..object import Object
from ..update import Update


class ShippingQuery(Object, Update):
    """This object contains information about an incoming shipping query.

    Parameters:
        id (``str``):
            Unique query identifier.

        from_user (:obj:`~Badmunda.types.User`):
            User who sent the query.

        invoice_payload (``str``):
            Bot-defined invoice payload, 1-128 bytes. This will not be displayed to the user, use for your internal processes.

        shipping_address (:obj:`~Badmunda.types.ShippingAddress`):
            User specified shipping address. Only available to the bot that received the payment.
    """

    def __init__(
        self,
        *,
        client: "Badmunda.Client" = None,
        id: str,
        from_user: "types.User",
        invoice_payload: str,
        shipping_address: Optional["types.ShippingAddress"] = None
    ):
        super().__init__(client)

        self.id = id
        self.from_user = from_user
        self.invoice_payload = invoice_payload
        self.shipping_address = shipping_address

    @staticmethod
    async def _parse(
        client: "Badmunda.Client",
        shipping_query: "raw.types.UpdateBotShippingQuery",
        users: dict
    ) -> "ShippingQuery":
        # Try to decode shipping query payload into string. If that fails, fallback to bytes instead of decoding by
        # ignoring/replacing errors, this way, button clicks will still work.
        try:
            payload = shipping_query.payload.decode()
        except (UnicodeDecodeError, AttributeError):
            payload = shipping_query.payload

        return ShippingQuery(
            id=str(shipping_query.query_id),
            from_user=types.User._parse(client, users[shipping_query.user_id]),
            invoice_payload=payload,
            shipping_address=types.ShippingAddress._parse(shipping_query.shipping_address),
            client=client
        )

    async def answer(
        self,
        ok: bool,
        shipping_options: "types.ShippingOptions" = None,
        error_message: str = None
    ):
        """Bound method *answer* of :obj:`~Badmunda.types.ShippingQuery`.

        Use this method as a shortcut for:

        .. code-block:: python

            await client.answer_shipping_query(
                shipping_query.id,
                ok=True
            )

        Example:
            .. code-block:: python

                await shipping_query.answer(ok=True)

        Parameters:
            ok (``bool``):
                Pass True if delivery to the specified address is possible and False if there are any problems (for example, if delivery to the specified address is not possible).

            shipping_options (:obj:`~Badmunda.types.ShippingOptions`, *optional*):
                Required if ok is True. A JSON-serialized array of available shipping options.

            error_message (``str``, *optional*):
                Required if ok is False. Error message in human readable form that explains why it is impossible to complete the order (e.g. "Sorry, delivery to your desired address is unavailable'). Telegram will display this message to the user.

        Returns:
            ``bool``: True, on success.
        """
        return await self._client.answer_shipping_query(
            shipping_query_id=self.id,
            ok=ok,
            shipping_options=shipping_options,
            error_message=error_message
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import List

import Badmunda
from Badmunda import raw, types


class GetSuitableDiscussionChats:
    async def get_suitable_discussion_chats(
        self: "Badmunda.Client"
    ) -> List["types.Chat"]:
        """Return a list of basic group and supergroup chats, which can be used as a discussion group for a channel.

        Returned basic group chats must be first upgraded to supergroups before they can be set as a discussion group.

        .. include:: /_includes/usable-by/users.rst

        Returns:
            List of :obj:`~Badmunda.types.Chat`: List of suitable discussion chats.

        Example:
            .. code-block:: python

                chats = await app.get_suitable_discussion_chats()
        """
        r = await self.invoke(
            raw.functions.channels.GetGroupsForDiscussion()
        )

        return types.List([types.Chat._parse_chat(self, i) for i in r.chats])



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from .apply_gift_code import ApplyGiftCode
from .check_gift_code import CheckGiftCode
from .convert_gift_to_stars import ConvertGiftToStars
from .get_available_gifts import GetAvailableGifts
from .get_chat_gifts import GetChatGifts
from .get_chat_gifts_count import GetChatGiftsCount
from .get_gift_upgrade_preview import GetGiftUpgradePreview
from .get_payment_form import GetPaymentForm
from .get_stars_balance import GetStarsBalance
from .get_upgraded_gift import GetUpgradedGift
from .hide_gift import HideGift
from .search_gifts_for_resale import SearchGiftsForResale
from .send_gift import SendGift
from .send_payment_form import SendPaymentForm
from .send_resold_gift import SendResoldGift
from .set_gift_resale_price import SetGiftResalePrice
from .set_pinned_gifts import SetPinnedGifts
from .show_gift import ShowGift
from .transfer_gift import TransferGift
from .upgrade_gift import UpgradeGift


class Payments(
    ApplyGiftCode,
    CheckGiftCode,
    ConvertGiftToStars,
    GetAvailableGifts,
    GetChatGifts,
    GetChatGiftsCount,
    GetGiftUpgradePreview,
    GetPaymentForm,
    GetStarsBalance,
    GetUpgradedGift,
    HideGift,
    SearchGiftsForResale,
    SendGift,
    SendPaymentForm,
    SendResoldGift,
    SetGiftResalePrice,
    SetPinnedGifts,
    ShowGift,
    TransferGift,
    UpgradeGift
):
    pass



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Optional

import Badmunda
from Badmunda import raw, types

from ..object import Object


class FormattedText(Object):
    """Contains information about a text with some entities.

    Parameters:
        text (``str``):
            The text.

        entities (List of :obj:`~Badmunda.types.MessageEntity`):
            Entities contained in the text. Entities can be nested, but must not mutually intersect with each other.
    """

    def __init__(
        self,
        *,
        text: str,
        entities: Optional[List["types.MessageEntity"]] = None,
    ):
        super().__init__()

        self.text = text
        self.entities = entities

    @staticmethod
    def _parse(client: "Badmunda.Client", text: "raw.types.TextWithEntities") -> "FormattedText":
        entities = types.List(
            filter(
                lambda x: x is not None,
                [types.MessageEntity._parse(client, entity, {}) for entity in text.entities]
            )
        )

        return FormattedText(
            text=text.text,
            entities=entities or None,
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Union

import Badmunda
from Badmunda import raw


class SetChatDiscussionGroup:
    async def set_chat_discussion_group(
        self: "Badmunda.Client", *,
        chat_id: Optional[Union[int, str]] = None,
        discussion_chat_id: Optional[Union[int, str]] = None
    ) -> bool:
        """Change the discussion group of a channel chat.

        Requires `can_change_info` administrator right in the channel if it is specified.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            chat_id (``int`` | ``str``, *optional*):
                Unique identifier (int) or username (str) of the target chat.
                Pass None to remove a link from the supergroup passed in the second argument to a linked channel chat (requires `can_pin_messages` member right in the supergroup)

            discussion_chat_id (``int`` | ``str``, *optional*):
                Unique identifier (int) or username (str) of a new channel's discussion group.
                Use the method :meth:`get_suitable_discussion_chats` to find all suitable groups.
                Pass None to remove the discussion group.

        Returns:
            ``bool``: True on success.

        Raises:
            ValueError: In case a chat id or discussion chat id belongs to a user.

        Example:
            .. code-block:: python

                # Set a discussion group for a channel
                await app.set_chat_discussion_group(chat_id="@Badmunda", discussion_chat_id="@Badmundachat")

                # Remove a discussion group from a channel
                await app.set_chat_discussion_group(chat_id="@Badmunda")

                # Remove a discussion group from a channel
                await app.set_chat_discussion_group(discussion_chat_id="@Badmundachat")
        """
        if chat_id is None and discussion_chat_id is None:
            raise ValueError("At least one of 'chat_id' or 'discussion_chat_id' must be provided.")

        if chat_id is None:
            channel_peer = raw.types.InputChannelEmpty()
        else:
            channel_peer = await self.resolve_peer(chat_id)

            if not isinstance(channel_peer, raw.types.InputPeerChannel):
                raise ValueError(f'The chat_id "{chat_id}" does not belong to a channel')

        if discussion_chat_id is None:
            discussion_peer = raw.types.InputChannelEmpty()
        else:
            discussion_peer = await self.resolve_peer(discussion_chat_id)

            if not isinstance(discussion_peer, raw.types.InputPeerChannel):
                raise ValueError(f'The discussion_chat_id "{discussion_chat_id}" does not belong to a chat')

        return bool(
            await self.invoke(
                raw.functions.channels.SetDiscussionGroup(
                    broadcast=channel_peer,
                    group=discussion_peer
                )
            )
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Optional

import Badmunda
from Badmunda import enums, raw, types, utils

from ..object import Object


class InputChecklistTask(Object):
    """Describes a task in a checklist to be sent.

    Parameters:
        id (``int``):
            Unique identifier of the task.

        text (``str``):
            Text of the task.

        parse_mode (:obj:`~Badmunda.enums.ParseMode`, *optional*):
            The parse mode to use for the checklist.

        entities (List of :obj:`~Badmunda.types.MessageEntity`, *optional*):
            List of special entities that appear in the checklist title.
    """

    def __init__(
        self,
        *,
        id: int,
        text: str,
        parse_mode: Optional["enums.ParseMode"] = None,
        entities: Optional[List["types.MessageEntity"]] = None,
    ):
        super().__init__()

        self.id = id
        self.text = text
        self.parse_mode = parse_mode
        self.entities = entities

    async def write(
        self, client: "Badmunda.Client"
    ) -> "raw.types.TodoItem":
        task_title, task_entities = (await utils.parse_text_entities(
            client, self.text, self.parse_mode, self.entities
        )).values()

        return raw.types.TodoItem(
            id=self.id,
            title=raw.types.TextWithEntities(
                text=task_title,
                entities=task_entities or []
            )
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from Badmunda import raw

from ..object import Object


class SavedCredentials(Object):
    """Contains information about saved payment credentials.

    Parameters:
        id (``str``):
            Unique identifier of the saved credentials.

        title (``str``):
            Title of the saved credentials.
    """
    def __init__(
        self,
        *,
        id: str,
        title: str
    ):
        super().__init__()

        self.id = id
        self.title = title

    @staticmethod
    def _parse(credential: "raw.base.PaymentSavedCredentials") -> "SavedCredentials":
        return SavedCredentials(
            id=credential.id,
            title=credential.title
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

import Badmunda
from Badmunda import raw


class AnswerPreCheckoutQuery:
    async def answer_pre_checkout_query(
        self: "Badmunda.Client",
        pre_checkout_query_id: str,
        ok: Optional[bool] = None,
        error_message: Optional[str] = None
    ):
        """Send answers to pre-checkout queries.

        .. include:: /_includes/usable-by/bots.rst

        Parameters:
            pre_checkout_query_id (``str``):
                Unique identifier for the query to be answered.

            ok (``bool``, *optional*):
                Set this flag if everything is alright (goods are available, etc.) and the bot is ready to proceed with the order.
                Otherwise do not set it, and set the error field, instead.

            error_message (``str``, *optional*):
                Error message in human readable form that explains the reason for failure to proceed with the checkout.
                Required if ``success`` isn't set.

        Returns:
            ``bool``: True, on success.

        Example:
            .. code-block:: python

                # Proceed with the order
                await app.answer_pre_checkout_query(query_id, ok=True)

                # Answer with error message
                await app.answer_pre_checkout_query(query_id, ok=False, error_message="Out of stock")
        """
        return await self.invoke(
            raw.functions.messages.SetBotPrecheckoutResults(
                query_id=int(pre_checkout_query_id),
                success=ok,
                error=error_message
            )
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import os
from typing import Union, BinaryIO, Callable

import Badmunda
from Badmunda import raw, types, utils, StopTransmission
from Badmunda.errors import FilePartMissing

class EditStoryMedia:
    async def edit_story_media(
        self: "Badmunda.Client",
        chat_id: Union[int, str],
        story_id: int,
        media: Union[str, BinaryIO] = None,
        duration: int = 0,
        width: int = 0,
        height: int = 0,
        thumb: Union[str, BinaryIO] = None,
        supports_streaming: bool = True,
        file_name: str = None,
        progress: Callable = None,
        progress_args: tuple = ()
    ) -> "types.Story":
        """Edit story media.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".

            story_id (``int``):
                Story identifier in the chat specified in chat_id.

            media (``str`` | ``BinaryIO``, *optional*):
                Video or photo to send.
                Pass a file_id as string to send a animation that exists on the Telegram servers,
                pass a file path as string to upload a new animation that exists on your local machine, or
                pass a binary file-like object with its attribute ".name" set for in-memory uploads.

            duration (``int``, *optional*):
                Duration of sent video in seconds.

            width (``int``, *optional*):
                Video width.

            height (``int``, *optional*):
                Video height.

            thumb (``str`` | ``BinaryIO``, *optional*):
                Thumbnail of the video sent.
                The thumbnail should be in JPEG format and less than 200 KB in size.
                A thumbnail's width and height should not exceed 320 pixels.
                Thumbnails can't be reused and can be only uploaded as a new file.

            progress (``Callable``, *optional*):
                Pass a callback function to view the file transmission progress.
                The function must take *(current, total)* as positional arguments (look at Other Parameters below for a
                detailed description) and will be called back each time a new file chunk has been successfully
                transmitted.

            progress_args (``tuple``, *optional*):
                Extra custom arguments for the progress callback function.
                You can pass anything you need to be available in the progress callback scope; for example, a Message
                object or a Client instance in order to edit the message with the updated progress status.

        Returns:
            :obj:`~Badmunda.types.Story`: On success, the edited story is returned.

        Example:
            .. code-block:: python

                # Replace the current media with a local photo
                await app.edit_story_media(chat_id, story_id, "new_photo.jpg")

                # Replace the current media with a local video
                await app.edit_story_media(chat_id, story_id, "new_video.mp4")
        """
        # TODO: media_areas

        try:
            if isinstance(media, str):
                if os.path.isfile(media):
                    thumb = await self.save_file(thumb)
                    file = await self.save_file(media, progress=progress, progress_args=progress_args)
                    mime_type = self.guess_mime_type(file.name)
                    if mime_type == "video/mp4":
                        media = raw.types.InputMediaUploadedDocument(
                            mime_type=mime_type,
                            file=file,
                            thumb=thumb,
                            attributes=[
                                raw.types.DocumentAttributeVideo(
                                    duration=duration,
                                    w=width,
                                    h=height,
                                ),
                                raw.types.DocumentAttributeFilename(file_name=file_name or os.path.basename(media))
                            ]
                        )
                    else:
                        media = raw.types.InputMediaUploadedPhoto(
                            file=file,
                        )
                else:
                    media = utils.get_input_media_from_file_id(media)
            else:
                thumb = await self.save_file(thumb)
                file = await self.save_file(media, progress=progress, progress_args=progress_args)
                mime_type = self.guess_mime_type(file.name)
                if mime_type == "video/mp4":
                    media = raw.types.InputMediaUploadedDocument(
                        mime_type=mime_type,
                        file=file,
                        thumb=thumb,
                        attributes=[
                            raw.types.DocumentAttributeVideo(
                                supports_streaming=supports_streaming or None,
                                duration=duration,
                                w=width,
                                h=height,
                            ),
                            raw.types.DocumentAttributeFilename(file_name=file_name or media.name)
                        ]
                    )
                else:
                    media = raw.types.InputMediaUploadedPhoto(
                        file=file,
                    )

            while True:
                try:
                    r = await self.invoke(
                        raw.functions.stories.EditStory(
                            peer=await self.resolve_peer(chat_id),
                            id=story_id,
                            media=media,
                        )
                    )
                except FilePartMissing as e:
                    await self.save_file(media, file_id=file.id, file_part=e.value)
                else:
                    for i in r.updates:
                        if isinstance(i, raw.types.UpdateStory):
                            return await types.Story._parse(
                                self,
                                i.story,
                                i.peer,
                                {i.id: i for i in r.users},
                                {i.id: i for i in r.chats}
                            )
        except StopTransmission:
            return None



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Optional, Union

from Badmunda import enums, types

from ..object import Object


class ReplyParameters(Object):
    """Describes reply parameters for the message that is being sent.

    Parameters:
        message_id  (``int``, *optional*):
            Identifier of the message that will be replied to in the current chat, or in the chat chat_id if it is specified.

        story_id  (``int``, *optional*):
            Unique identifier for the story in the chat.

        chat_id (``int`` | ``str``, *optional*):
            Unique identifier (int) or username (str) of the target chat.
            For your personal cloud (Saved Messages) you can simply use "me" or "self".
            For a contact that exists in your Telegram address book you can use his phone number (str).

        quote (``str``, *optional*):
            Quoted part of the message to be replied to, 0-1024 characters after entities parsing.
            The quote must be an exact substring of the message to be replied to, including bold, italic, underline, strikethrough, spoiler, and custom_emoji entities.
            The message will fail to send if the quote isn't found in the original message.

        quote_parse_mode (:obj:`~Badmunda.enums.ParseMode`, *optional*):
            By default, texts are parsed using both Markdown and HTML styles.
            You can combine both syntaxes together.

        quote_entities (List of :obj:`~Badmunda.types.MessageEntity`, *optional*):
            List of special entities that appear in message text, which can be specified instead of *quote_parse_mode*.

        quote_position (``int``, *optional*):
            Position of the quote in the original message in UTF-16 code units.
    """

    def __init__(
        self,
        *,
        message_id: Optional[int] = None,
        story_id: Optional[int] = None,
        chat_id: Optional[Union[int, str]] = None,
        quote: Optional[str] = None,
        quote_parse_mode: Optional["enums.ParseMode"] = None,
        quote_entities: Optional[List["types.MessageEntity"]] = None,
        quote_position: Optional[int] = None
    ):
        super().__init__()

        self.message_id = message_id
        self.story_id = story_id
        self.chat_id = chat_id
        self.quote = quote
        self.quote_parse_mode = quote_parse_mode
        self.quote_entities = quote_entities
        self.quote_position = quote_position



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

from Badmunda import raw, types
from ..object import Object


class BotVerification(Object):
    """Information about bot verification.

    Parameters:
        bot (:obj:`~Badmunda.types.User`):
            Bot that is verified this user.

        custom_emoji_id (``int``):
            Custom emoji icon identifier.

        description (``int``, *optional*):
            Additional description about the verification.
    """

    def __init__(
        self,
        *,
        bot: int,
        custom_emoji_id: int,
        description: str
    ):
        self.bot = bot
        self.custom_emoji_id = custom_emoji_id
        self.description = description

    @staticmethod
    def _parse(
        client,
        verification: "raw.types.BotVerification",
        users
    ) -> Optional["BotVerification"]:
        if not verification:
            return None

        return BotVerification(
            bot=types.User._parse(client, users.get(verification.bot_id)),
            custom_emoji_id=verification.icon,
            description=verification.description
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

import Badmunda

from Badmunda import raw
from ..object import Object



class GiveawayCreated(Object):
    """This object represents a service message about the creation of a scheduled giveaway.

    Parameters:
        prize_star_count (``int``, *optional*):
            The number of Telegram Stars to be split between giveaway winners.
            For Telegram Star giveaways only.
    """

    def __init__(
        self,
        *,
        client: "Badmunda.Client" = None,
        prize_star_count: Optional[int] = None
    ):
        super().__init__(client)

        self.prize_star_count = prize_star_count


    @staticmethod
    def _parse(
        client,
        giveaway_launch: "raw.types.MessageActionGiveawayLaunch"
    ) -> "GiveawayCreated":
        if isinstance(giveaway_launch, raw.types.MessageActionGiveawayLaunch):
            return GiveawayCreated(
                client=client,
                prize_star_count=getattr(giveaway_launch, "stars", None)
            )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from .animation import Animation
from .audio import Audio
from .available_effect import AvailableEffect
from .boosts_status import BoostsStatus
from .business_message import BusinessMessage
from .chat_background import ChatBackground
from .chat_boost import ChatBoost
from .chat_theme import ChatTheme
from .checked_gift_code import CheckedGiftCode
from .checklist_task import ChecklistTask
from .checklist_tasks_added import ChecklistTasksAdded
from .checklist_tasks_done import ChecklistTasksDone
from .checklist import Checklist
from .contact_registered import ContactRegistered
from .contact import Contact
from .dice import Dice
from .direct_message_price_changed import DirectMessagePriceChanged
from .direct_topic import DirectMessagesTopic
from .document import Document
from .external_reply_info import ExternalReplyInfo
from .fact_check import FactCheck
from .formatted_text import FormattedText
from .forum_topic import ForumTopic
from .forum_topic_closed import ForumTopicClosed
from .forum_topic_created import ForumTopicCreated
from .forum_topic_edited import ForumTopicEdited
from .forum_topic_reopened import ForumTopicReopened
from .game import Game
from .general_forum_topic_hidden import GeneralForumTopicHidden
from .general_forum_topic_unhidden import GeneralForumTopicUnhidden
from .gift_code import GiftCode
from .gift_upgrade_preview import GiftUpgradePreview
from .invoice import Invoice
from .link_preview_options import LinkPreviewOptions
from .giveaway import Giveaway
from .input_checklist_task import InputChecklistTask
from .giveaway_completed import GiveawayCompleted
from .giveaway_created import GiveawayCreated
from .giveaway_prize_stars import GiveawayPrizeStars
from .giveaway_winners import GiveawayWinners
from .location import Location
from .mask_position import MaskPosition
from .media_area import MediaArea
from .message import Message
from .message_entity import MessageEntity
from .message_origin import MessageOrigin
from .message_origin_channel import MessageOriginChannel
from .message_origin_chat import MessageOriginChat
from .message_origin_hidden_user import MessageOriginHiddenUser
from .message_origin_import import MessageOriginImport
from .message_origin_user import MessageOriginUser
from .message_reactions import MessageReactions
from .my_boost import MyBoost
from .paid_media_info import PaidMediaInfo
from .paid_media_preview import PaidMediaPreview
from .paid_messages_price_changed import PaidMessagesPriceChanged
from .paid_messages_refunded import PaidMessagesRefunded
from .paid_reactor import PaidReactor
from .payment_form import PaymentForm
from .payment_option import PaymentOption
from .payment_result import PaymentResult
from .photo import Photo
from .poll import Poll
from .proximity_alert_triggered import ProximityAlertTriggered
from .poll_option import PollOption
from .reaction import Reaction
from .refunded_payment import RefundedPayment
from .reply_parameters import ReplyParameters
from .restriction_reason import RestrictionReason
from .saved_credentials import SavedCredentials
from .screenshot_taken import ScreenshotTaken
from .gift_attribute import GiftAttribute
from .gift import Gift
from .gifted_premium import GiftedPremium
from .gifted_stars import GiftedStars
from .sticker import Sticker
from .story_view import StoryView
from .story import Story
from .stripped_thumbnail import StrippedThumbnail
from .successful_payment import SuccessfulPayment
from .text_quote import TextQuote
from .thumbnail import Thumbnail
from .upgraded_gift_attribute_id_backdrop import UpgradedGiftAttributeIdBackdrop
from .upgraded_gift_attribute_id_model import UpgradedGiftAttributeIdModel
from .upgraded_gift_attribute_id_symbol import UpgradedGiftAttributeIdSymbol
from .upgraded_gift_attribute_id import UpgradedGiftAttributeId
from .venue import Venue
from .video import Video
from .video_note import VideoNote
from .voice import Voice
from .web_app_data import WebAppData
from .web_page import WebPage
from .write_access_allowed import WriteAccessAllowed

__all__ = [
    "Animation",
    "Audio",
    "AvailableEffect",
    "BoostsStatus",
    "BusinessMessage",
    "ChatBackground",
    "ChatBoost",
    "ChatTheme",
    "CheckedGiftCode",
    "ChecklistTask",
    "ChecklistTasksAdded",
    "ChecklistTasksDone",
    "Checklist",
    "ContactRegistered",
    "Contact",
    "Dice",
    "DirectMessagePriceChanged",
    "DirectMessagesTopic",
    "Document",
    "ExternalReplyInfo",
    "FactCheck",
    "FormattedText",
    "ForumTopic",
    "ForumTopicClosed",
    "ForumTopicCreated",
    "ForumTopicEdited",
    "ForumTopicReopened",
    "Game",
    "GeneralForumTopicHidden",
    "GeneralForumTopicUnhidden",
    "GiftCode",
    "GiftUpgradePreview",
    "Giveaway",
    "InputChecklistTask",
    "Invoice",
    "LinkPreviewOptions",
    "GiveawayCompleted",
    "GiveawayCreated",
    "GiveawayPrizeStars",
    "GiveawayWinners",
    "Location",
    "MaskPosition",
    "MediaArea",
    "Message",
    "MessageEntity",
    "MessageOrigin",
    "MessageOriginChannel",
    "MessageOriginChat",
    "MessageOriginHiddenUser",
    "MessageOriginImport",
    "MessageOriginUser",
    "MessageReactions",
    "MyBoost",
    "PaidMediaInfo",
    "PaidMediaPreview",
    "PaidMessagesPriceChanged",
    "PaidMessagesRefunded",
    "PaidReactor",
    "PaymentForm",
    "PaymentOption",
    "PaymentResult",
    "Photo",
    "Poll",
    "ProximityAlertTriggered",
    "PollOption",
    "Reaction",
    "RefundedPayment",
    "ReplyParameters",
    "RestrictionReason",
    "SavedCredentials",
    "ScreenshotTaken",
    "GiftAttribute",
    "Gift",
    "GiftedPremium",
    "GiftedStars",
    "Sticker",
    "StoryView",
    "Story",
    "StrippedThumbnail",
    "SuccessfulPayment",
    "TextQuote",
    "Thumbnail",
    "UpgradedGiftAttributeIdBackdrop",
    "UpgradedGiftAttributeIdModel",
    "UpgradedGiftAttributeIdSymbol",
    "UpgradedGiftAttributeId",
    "Venue",
    "Video",
    "VideoNote",
    "Voice",
    "WebAppData",
    "WebPage",
    "WriteAccessAllowed",
]



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import logging
from typing import Optional, Type

from .transport import TCP, TCPAbridged
from ..session.internals import DataCenter

log = logging.getLogger(__name__)


class Connection:
    MAX_CONNECTION_ATTEMPTS = 3

    def __init__(
        self,
        dc_id: int,
        test_mode: bool,
        ipv6: bool,
        proxy: dict,
        media: bool = False,
        protocol_factory: Type[TCP] = TCPAbridged,
        loop: Optional[asyncio.AbstractEventLoop] = None
    ) -> None:
        self.dc_id = dc_id
        self.test_mode = test_mode
        self.ipv6 = ipv6
        self.proxy = proxy
        self.media = media
        self.protocol_factory = protocol_factory

        self.address = DataCenter(dc_id, test_mode, ipv6, media)
        self.protocol: Optional[TCP] = None

        if isinstance(loop, asyncio.AbstractEventLoop):
            self.loop = loop
        else:
            self.loop = asyncio.get_event_loop()

    async def connect(self) -> None:
        for i in range(Connection.MAX_CONNECTION_ATTEMPTS):
            self.protocol = self.protocol_factory(ipv6=self.ipv6, proxy=self.proxy, loop=self.loop)

            try:
                log.info("Connecting...")
                await self.protocol.connect(self.address)
            except OSError as e:
                log.warning("Unable to connect due to network issues: %s", e)
                await self.protocol.close()
                await asyncio.sleep(1)
            else:
                log.info("Connected! %s DC%s%s - IPv%s",
                         "Test" if self.test_mode else "Production",
                         self.dc_id,
                         " (media)" if self.media else "",
                         "6" if self.ipv6 else "4")
                break
        else:
            log.warning("Connection failed! Trying again...")
            raise ConnectionError

    async def close(self) -> None:
        await self.protocol.close()
        log.info("Disconnected")

    async def send(self, data: bytes) -> None:
        await self.protocol.send(data)

    async def recv(self) -> Optional[bytes]:
        return await self.protocol.recv()



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, Optional, Union

import Badmunda
from Badmunda import enums, raw, types, utils

from ..object import Object


class ChatShared(Object):
    """Contains information about a chat shared with a bot.

    Parameters:
        button_id (``int``):
            Identifier of button.

        chat (:obj:`~Badmunda.types.Chat`):
            Requested chats.
    """
    def __init__(
        self, *,
        button_id: int,
        chat: "types.Chat",
    ):
        super().__init__()

        self.button_id = button_id
        self.chat = chat

    @staticmethod
    def _parse(
        client: "Badmunda.Client",
        action: Union[
            "raw.types.MessageActionRequestedPeer",
            "raw.types.MessageActionRequestedPeerSentMe"
        ],
        chats: Dict[int, "raw.base.Chat"] = {}
    ) -> Optional["ChatShared"]:
        peer = action.peers[0]

        if isinstance(peer, (raw.types.PeerUser, raw.types.RequestedPeerUser)):
            return None

        peer_id = utils.get_peer_id(peer)
        peer_type = utils.get_peer_type(peer_id)

        if peer_type == "chat":
            chat_type = enums.ChatType.GROUP
        else:
            chat_type = enums.ChatType.CHANNEL

        chat_shared = None

        if isinstance(action, raw.types.MessageActionRequestedPeer):
            raw_chat = chats.get(utils.get_raw_peer_id(peer))

            if raw_chat:
                chat_shared = types.Chat._parse_chat(client, raw_chat)
            else:
                chat_shared = types.Chat(
                    id=peer_id,
                    type=chat_type,
                    client=client
                )
        elif isinstance(action, raw.types.MessageActionRequestedPeerSentMe):
            chat_shared = types.Chat(
                id=peer_id,
                type=chat_type,
                first_name=getattr(peer, "first_name", None),
                last_name=getattr(peer, "last_name", None),
                title=getattr(peer, "title", None),
                username=getattr(peer, "username", None),
                photo=types.Photo._parse(client, getattr(peer, "photo", None)),
                client=client
            )

        return ChatShared(
            button_id=action.button_id,
            chat=chat_shared
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from typing import Union, List, Iterable

import Badmunda
from Badmunda import raw
from Badmunda import types


class GetUsers:
    async def get_users(
        self: "Badmunda.Client",
        user_ids: Union[int, str, Iterable[Union[int, str]]]
    ) -> Union["types.User", List["types.User"]]:
        """Get information about a user.
        You can retrieve up to 200 users at once.

        .. include:: /_includes/usable-by/users-bots.rst

        Parameters:
            user_ids (``int`` | ``str`` | Iterable of ``int`` or ``str``):
                A list of User identifiers (id or username) or a single user id/username.
                For a contact that exists in your Telegram address book you can use his phone number (str).

        Returns:
            :obj:`~Badmunda.types.User` | List of :obj:`~Badmunda.types.User`: In case *user_ids* was not a list,
            a single user is returned, otherwise a list of users is returned.

        Example:
            .. code-block:: python

                # Get information about one user
                await app.get_users("me")

                # Get information about multiple users at once
                await app.get_users([user_id1, user_id2, user_id3])
        """

        is_iterable = not isinstance(user_ids, (int, str))
        user_ids = list(user_ids) if is_iterable else [user_ids]
        user_ids = await asyncio.gather(*[self.resolve_peer(i) for i in user_ids])

        r = await self.invoke(
            raw.functions.users.GetUsers(
                id=user_ids
            )
        )

        users = types.List()

        for i in r:
            users.append(types.User._parse(self, i))

        return users if is_iterable else users[0]



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Optional

import Badmunda
from Badmunda import enums, raw, types

from ..object import Object


class PaymentForm(Object):
    """Contains information about an invoice payment form.

    Parameters:
        id (``int``):
            Form id.

        type (:obj:`~Badmunda.enums.PaymentFormType`):
            Type of the payment form.

        title (``str``, *optional*):
            Form title.

        description (``str``, *optional*):
            Form description.

        photo (:obj:`~Badmunda.types.Photo`, *optional*):
            Product photo.

        seller_bot_user_id (``int``, *optional*):
            User identifier of the seller bot.

        seller_bot (:obj:`~Badmunda.types.User`, *optional*):
            Information about the seller bot.

        payment_provider_user_id (``int``, *optional*):
            User identifier of the payment provider bot.

        payment_provider (:obj:`~Badmunda.types.User`, *optional*):
            Information about the payment provider.

        additional_payment_options (List of :obj:`~Badmunda.types.PaymentOption`, *optional*):
            The list of additional payment options.

        saved_credentials (List of :obj:`~Badmunda.types.SavedCredentials`, *optional*):
            The list of saved payment credentials.

        invoice (:obj:`~Badmunda.types.Invoice`, *optional*):
            Full information about the invoice.

        url (``str``, *optional*):
            Payment form URL.

        can_save_credentials (``bool``, *optional*):
            True, if the user can choose to save credentials.

        need_password (``bool``, *optional*):
            True, if the user will be able to save credentials, if sets up a 2-step verification password.

        native_provider (``str``, *optional*):
            Payment provider name.

        raw (:obj:`~raw.base.payments.PaymentForm`, *optional*):
            The raw object, as received from the Telegram API.
    """

    def __init__(
        self,
        *,
        client: "Badmunda.Client" = None,
        id: int,
        type: "enums.PaymentFormType",
        title: Optional[str] = None,
        description: Optional[str] = None,
        photo: Optional["types.Photo"] = None,
        seller_bot_user_id: Optional[int] = None,
        seller_bot: Optional["types.User"] = None,
        payment_provider_user_id: Optional[int] = None,
        payment_provider: Optional["types.User"] = None,
        additional_payment_options: Optional[List["types.PaymentOption"]] = None,
        saved_credentials: Optional[List["types.SavedCredentials"]] = None,
        invoice: Optional["types.Invoice"] = None,
        url: Optional[str] = None,
        can_save_credentials: Optional[bool] = None,
        need_password: Optional[bool] = None,
        native_provider: Optional[str] = None,
        raw: "raw.base.payments.PaymentForm" = None,
    ):
        super().__init__(client)

        self.id = id
        self.type = type
        self.seller_bot_user_id = seller_bot_user_id
        self.seller_bot = seller_bot
        self.payment_provider_user_id = payment_provider_user_id
        self.payment_provider = payment_provider
        self.additional_payment_options = additional_payment_options
        self.saved_credentials = saved_credentials
        self.title = title
        self.description = description
        self.photo = photo
        self.invoice = invoice
        self.url = url
        self.can_save_credentials = can_save_credentials
        self.need_password = need_password
        self.native_provider = native_provider
        self.raw = raw

    @staticmethod
    def _parse(client, form: "raw.base.payments.PaymentForm") -> "PaymentForm":
        users = {i.id: i for i in getattr(form, "users", [])}

        if isinstance(form, raw.types.payments.PaymentForm):
            return PaymentForm(
                id=form.form_id,
                type=enums.PaymentFormType.REGULAR,
                title=form.title,
                description=form.description,
                photo=types.Photo._parse(client, form.photo),
                seller_bot_user_id=form.bot_id,
                seller_bot=types.User._parse(client, users.get(form.bot_id)),
                payment_provider_user_id=form.provider_id,
                payment_provider=types.User._parse(client, users.get(form.provider_id)),
                invoice=types.Invoice._parse(client, form.invoice),
                url=form.url,
                can_save_credentials=form.can_save_credentials,
                need_password=form.password_missing,
                native_provider=form.native_provider,
                # native_params,
                additional_payment_options=types.List([types.PaymentOption._parse(option) for option in getattr(form, "additional_methods", [])]) or None,
                # saved_info,
                saved_credentials=types.List([types.SavedCredentials._parse(credential) for credential in getattr(form, "saved_credentials", [])]) or None,
                raw=form
            )
        elif isinstance(form, raw.types.payments.PaymentFormStarGift):
            return PaymentForm(
                id=form.form_id,
                type=enums.PaymentFormType.STAR_SUBSCRIPTION,
                invoice=types.Invoice._parse(client, form.invoice),
                raw=form
            )
        elif isinstance(form, raw.types.payments.PaymentFormStars):
            return PaymentForm(
                id=form.form_id,
                type=enums.PaymentFormType.STARS,
                title=form.title,
                description=form.description,
                photo=types.Photo._parse(client, form.photo),
                seller_bot_user_id=form.bot_id,
                seller_bot=types.User._parse(client, users.get(form.bot_id)),
                invoice=types.Invoice._parse(client, form.invoice),
                raw=form
            )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List, Optional

import Badmunda
from Badmunda import raw, types

from ..object import Object


class MessageReactions(Object):
    """Contains information about a message reactions.

    Parameters:
        reactions (List of :obj:`~Badmunda.types.Reaction`):
            Reactions list.

        are_tags (``bool``, *optional*):
            True, if the reactions are tags and Telegram Premium users can filter messages by them.

        paid_reactors (List of :obj:`~Badmunda.types.PaidReactor`, *optional*):
            Information about top users that added the paid reaction.

        can_get_added_reactions (``bool``, *optional*):
            True, if the list of added reactions is available using :meth:`~Badmunda.Client.get_message_added_reactions`.
    """

    # TODO: Add get_message_added_reactions method

    def __init__(
        self,
        *,
        client: "Badmunda.Client" = None,
        reactions: Optional[List["types.Reaction"]] = None,
        are_tags: Optional[bool] = None,
        paid_reactors: Optional[List["types.PaidReactor"]] = None,
        can_get_added_reactions: Optional[bool] = None,
    ):
        super().__init__(client)

        self.reactions = reactions
        self.are_tags = are_tags
        self.paid_reactors = paid_reactors
        self.can_get_added_reactions = can_get_added_reactions

    @staticmethod
    def _parse(
        client: "Badmunda.Client",
        message_reactions: Optional["raw.base.MessageReactions"],
        users: Dict[int, "types.User"],
        chats: Dict[int, "types.Chat"],
    ) -> Optional["MessageReactions"]:
        if not message_reactions:
            return None

        return MessageReactions(
            client=client,
            reactions=types.List(
                [
                    types.Reaction._parse_count(client, reaction)
                    for reaction in message_reactions.results
                ]
            ),
            are_tags=message_reactions.reactions_as_tags,
            paid_reactors=types.List(
                [
                    types.PaidReactor._parse(client, paid_reactor, users, chats)
                    for paid_reactor in message_reactions.top_reactors
                ]
            )
            if message_reactions.top_reactors
            else None,
            can_get_added_reactions=message_reactions.can_see_list,
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, List, Union

import Badmunda
from Badmunda import raw, types
from ..object import Object


class Invoice(Object):
    """This object contains basic information about an invoice.

    Parameters:
        currency (``str``):
            Three-letter ISO 4217 `currency <https://core.telegram.org/bots/payments#supported-currencies>`_ code.

        is_test (``bool``):
            True, if the invoice is a test invoice.

        title (``str``, *optional*):
            Product name.

        description (``str``, *optional*):
            Product description.

        total_amount (``int``, *optional*):
            Total price in the smallest units of the currency (integer, **not** float/double). For example, for a price of ``US$ 1.45`` pass ``amount = 145``. See the exp parameter in `currencies.json <https://core.telegram.org/bots/payments/currencies.json>`_, it shows the number of digits past the decimal point for each currency (2 for the majority of currencies).

        start_parameter (``str``, *optional*):
            Unique bot deep-linking parameter that can be used to generate this invoice.

        prices (List of :obj:`~Badmunda.types.LabeledPrice`, *optional*):
            Price breakdown, a list of components (e.g. product price, tax, discount, delivery cost, delivery tax, bonus, etc.).

        is_name_requested (``bool``, *optional*):
            True, if the name should be specified.

        is_phone_requested (``bool``, *optional*):
            True, if the phone should be specified.

        is_email_requested (``bool``, *optional*):
            True, if the email address should be specified.

        is_shipping_address_requested (``bool``, *optional*):
            True, if the shipping address should be specified.

        is_flexible (``bool``, *optional*):
            True, if the final price depends on the shipping method.

        is_phone_to_provider (``bool``, *optional*):
            True, if user's phone should be sent to provider.

        is_email_to_provider (``bool``, *optional*):
            True, if user's email address should be sent to provider.

        is_recurring (``bool``, *optional*):
            Whether this is a recurring payment.

        max_tip_amount (``int``, *optional*):
            The maximum accepted amount for tips in the smallest units of the currency (integer, not float/double).
            For example, for a price of US$ 1.45 pass amount = 145.
            See the exp parameter in currencies.json, it shows the number of digits past the decimal point for each currency (2 for the majority of currencies).

        suggested_tip_amounts (List of ``int``, *optional*):
            A vector of suggested amounts of tips in the smallest units of the currency (integer, not float/double).
            At most 4 suggested tip amounts can be specified.
            The suggested tip amounts must be positive, passed in a strictly increased order and must not exceed max_tip_amount.

        terms_url (``str``, *optional*):
            Terms of service URL.

        raw (:obj:`~raw.base.payments.MessageMediaInvoice` | :obj:`~raw.base.Invoice`, *optional*):
            The raw object, as received from the Telegram API.
    """

    def __init__(
        self,
        *,
        client: "Badmunda.Client" = None,
        currency: str,
        is_test: bool,
        title: Optional[str] = None,
        description: Optional[str] = None,
        total_amount: Optional[int] = None,
        start_parameter: Optional[str] = None,
        prices: Optional[List["types.LabeledPrice"]] = None,
        is_name_requested: Optional[bool] = None,
        is_phone_requested: Optional[bool] = None,
        is_email_requested: Optional[bool] = None,
        is_shipping_address_requested: Optional[bool] = None,
        is_flexible: Optional[bool] = None,
        is_phone_to_provider: Optional[bool] = None,
        is_email_to_provider: Optional[bool] = None,
        is_recurring: Optional[bool] = None,
        max_tip_amount: Optional[int] = None,
        suggested_tip_amounts: Optional[List[int]] = None,
        terms_url: Optional[str] = None,
        raw: Union["raw.types.MessageMediaInvoice", "raw.types.Invoice"] = None
    ):
        super().__init__(client)

        self.currency = currency
        self.is_test = is_test
        self.title = title
        self.description = description
        self.total_amount = total_amount
        self.start_parameter = start_parameter
        self.prices = prices
        self.is_name_requested = is_name_requested
        self.is_phone_requested = is_phone_requested
        self.is_email_requested = is_email_requested
        self.is_shipping_address_requested = is_shipping_address_requested
        self.is_flexible = is_flexible
        self.is_phone_to_provider = is_phone_to_provider
        self.is_email_to_provider = is_email_to_provider
        self.is_recurring = is_recurring
        self.max_tip_amount = max_tip_amount
        self.suggested_tip_amounts = suggested_tip_amounts
        self.terms_url = terms_url
        self.raw = raw

    @staticmethod
    def _parse(client, invoice: Union["raw.types.MessageMediaInvoice", "raw.types.Invoice"]) -> "Invoice":
        return Invoice(
            currency=invoice.currency,
            is_test=invoice.test,
            title=getattr(invoice, "title", None),
            description=getattr(invoice, "description", None),
            total_amount=getattr(invoice, "total_amount", None),
            start_parameter=getattr(invoice, "start_param", None) or None,
            prices=types.List(types.LabeledPrice._parse(lp) for lp in invoice.prices) if getattr(invoice, "prices", None) else None,
            is_name_requested=getattr(invoice, "name_requested", None),
            is_phone_requested=getattr(invoice, "phone_requested", None),
            is_email_requested=getattr(invoice, "email_requested", None),
            is_shipping_address_requested=getattr(invoice, "shipping_address_requested", None),
            is_flexible=getattr(invoice, "flexible", None),
            is_phone_to_provider=getattr(invoice, "phone_to_provider", None),
            is_email_to_provider=getattr(invoice, "email_to_provider", None),
            is_recurring=getattr(invoice, "recurring", None),
            max_tip_amount=getattr(invoice, "max_tip_amount", None),
            suggested_tip_amounts=getattr(invoice, "suggested_tip_amounts", None) or None,
            terms_url=getattr(invoice, "terms_url", None),
            raw=invoice,
            client=client
            # TODO: Add photo and extended media
        )



#  Badmunda - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Badmunda.
#
#  Badmunda is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Badmunda is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Badmunda.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from typing import List, Optional

import Badmunda
from Badmunda import raw, utils
from Badmunda import types
from Badmunda.file_id import FileId, FileType, FileUniqueId, FileUniqueType, ThumbnailSource
from ..object import Object


class Animation(Object):
    """An animation file (GIF or H.264/MPEG-4 AVC video without sound).

    Parameters:
        file_id (``str``):
            Identifier for this file, which can be used to download or reuse the file.

        file_unique_id (``str``):
            Unique identifier for this file, which is supposed to be the same over time and for different accounts.
            Can't be used to download or reuse the file.

        width (``int``):
            Animation width as defined by sender.

        height (``int``):
            Animation height as defined by sender.

        duration (``int``):
            Duration of the animation in seconds as defined by sender.

        file_name (``str``, *optional*):
            Animation file name.

        mime_type (``str``, *optional*):
            Mime type of a file as defined by sender.

        file_size (``int``, *optional*):
            File size.

        date (:py:obj:`~datetime.datetime`, *optional*):
            Date the animation was sent.

        thumbs (List of :obj:`~Badmunda.types.Thumbnail`, *optional*):
            Animation thumbnails.
    """

    def __init__(
        self,
        *,
        client: "Badmunda.Client" = None,
        file_id: str,
        file_unique_id: str,
        width: int,
        height: int,
        duration: int,
        file_name: str = None,
        mime_type: str = None,
        file_size: int = None,
        date: datetime = None,
        thumbs: List["types.Thumbnail"] = None
    ):
        super().__init__(client)

        self.file_id = file_id
        self.file_unique_id = file_unique_id
        self.file_name = file_name
        self.mime_type = mime_type
        self.file_size = file_size
        self.date = date
        self.width = width
        self.height = height
        self.duration = duration
        self.thumbs = thumbs

    async def add_to_gifs(
        self,
        unsave: bool = False
    ) -> bool:
        """Bound method *add_to_gifs* of :obj:`~Badmunda.types.Message`.

        .. include:: /_includes/usable-by/users.rst

        Use as a shortcut for:

        .. code-block:: python

            await app.add_to_gifs(message.animation.file_id)

        Parameters:
            unsave (``bool``, optional):
                Whether to remove the GIF from the list of saved GIFs. Defaults to ``False``.

        Returns:
            ``bool``: True on success.
        """
        return await self._client.add_to_gifs(self.file_id, unsave)

    @staticmethod
    def _parse(
        client,
        animation: "raw.types.Document",
        video_attributes: "raw.types.DocumentAttributeVideo",
        file_name: str
    ) -> "Animation":
        return Animation(
            file_id=FileId(
                file_type=FileType.ANIMATION,
                dc_id=animation.dc_id,
                media_id=animation.id,
                access_hash=animation.access_hash,
                file_reference=animation.file_reference
            ).encode(),
            file_unique_id=FileUniqueId(
                file_unique_type=FileUniqueType.DOCUMENT,
                media_id=animation.id
            ).encode(),
            width=getattr(video_attributes, "w", 0),
            height=getattr(video_attributes, "h", 0),
            duration=getattr(video_attributes, "duration", 0),
            mime_type=animation.mime_type,
            file_size=animation.size,
            file_name=file_name,
            date=utils.timestamp_to_datetime(animation.date),
            thumbs=types.Thumbnail._parse(client, animation),
            client=client
        )

    @staticmethod
    def _parse_chat_animation(
        client,
        video: "raw.types.Photo",
        file_name: str
    ) -> Optional["Animation"]:
        if isinstance(video, raw.types.Photo):
            if not video.video_sizes:
                return None

            videos: List[raw.types.VideoSize] = []

            for v in video.video_sizes:
                if isinstance(v, raw.types.VideoSize):
                    videos.append(v)

            videos.sort(key=lambda v: v.w * v.h)

            main = videos[-1]

            return Animation(
                file_id=FileId(
                    file_type=FileType.PHOTO,
                    dc_id=video.dc_id,
                    media_id=video.id,
                    access_hash=video.access_hash,
                    file_reference=video.file_reference,
                    thumbnail_source=ThumbnailSource.THUMBNAIL,
                    thumbnail_file_type=FileType.PHOTO,
                    thumbnail_size=main.type,
                    volume_id=0,
                    local_id=0
                ).encode(),
                file_unique_id=FileUniqueId(
                    file_unique_type=FileUniqueType.DOCUMENT,
                    media_id=video.id
                ).encode(),
                width=main.w,
                height=main.h,
                duration=0,
                file_size=main.size,
                date=utils.timestamp_to_datetime(video.date),
                file_name=file_name,
                mime_type="video/mp4",
                client=client
            )
